""" betterAsteroids.py 
    Show more features of game engine
    demonstrates multiple states, multiple
    sprite groups, carrier object
"""

import pygame, gameEngine, random

class Ship(gameEngine.SuperSprite):
    """ player avatar.
        standard space controls
        arrows to turn, accel, space
        to fire
    """
    def __init__(self, scene):
        gameEngine.SuperSprite.__init__(self, scene)
        self.setImage("ship.gif")
        self.setSpeed(0)
        self.setAngle(0)
        
    def checkEvents(self):
        keys = pygame.key.get_pressed()
        if keys[pygame.K_LEFT]:
            self.rotateBy(5)
        if keys[pygame.K_RIGHT]:
            self.rotateBy(-5)
        if keys[pygame.K_UP]:
            self.addForce(.2, self.rotation)
        if keys[pygame.K_SPACE]:
            self.scene.bullet.fire()
    
class Bullet(gameEngine.SuperSprite):
    """ bullet fired from spacecraft 
    """
    
    def __init__(self, scene):
        gameEngine.SuperSprite.__init__(self, scene)
        self.setImage("bullet.gif")
        self.imageMaster = pygame.transform.scale(self.imageMaster, (5, 5))
        self.setBoundAction(self.HIDE)
        self.reset()
        
    def fire(self):
        self.setPosition((self.scene.ship.x, self.scene.ship.y))
        self.setSpeed(12)
        self.setAngle(self.scene.ship.rotation)
        
    def reset(self):
        self.setPosition ((-100, -100))
        self.setSpeed(0)
    
class Rock(gameEngine.SuperSprite):
    """ asteroid.  Rotates, wraps, and generally
        gets in the way
    """
    def __init__(self, scene):
        gameEngine.SuperSprite.__init__(self, scene)
        self.setImage("rock.gif")
        self.reset()
        
    def checkEvents(self):
        self.rotateBy(self.rotSpeed)
        
    def reset(self):
        """ change attributes randomly """
        
        #set random position
        x = random.randint(0, self.screen.get_width())
        y = random.randint(0, self.screen.get_height())
        self.setPosition((x, y))
        
        #set random size
        scale = random.randint(10, 40)
        self.setImage("rock.gif")
        self.imageMaster = \
            pygame.transform.scale(self.imageMaster, (scale, scale))
        
        self.setSpeed(random.randint(0, 6))
        self.setAngle(random.randint(0, 360))
        self.rotSpeed = random.randint(-5, 5)

class Game(gameEngine.Scene):
    """ primary gameplay state
        manages player, bullet, rocks
        returns score in carrier object
    """
    
    def __init__(self, carrier):
        gameEngine.Scene.__init__(self)
        self.carrier = carrier
        
        self.ship = Ship(self)
        self.bullet = Bullet(self)
        self.numRocks = 10
        self.points = 2000
        
        self.lblPoints = gameEngine.Label()
        self.lblPoints.center = (100, 30)
        self.lblPoints.text = "%d" % self.points
        self.lblPoints.fgColor = (0xFF, 0xFF, 0xFF)
        self.lblPoints.bgColor = (0, 0, 0)
        self.rocks = []
        for i in range(self.numRocks):
            self.rocks.append(Rock(self))
            
        self.rockGroup = self.makeSpriteGroup(self.rocks)
        self.addGroup(self.rockGroup)
        self.sprites = [self.lblPoints, self.bullet, self.ship ]
        self.setCaption("asteroids")       
        
    def update(self):
        rockHitShip = self.ship.collidesGroup(self.rocks)
        if rockHitShip:
            rockHitShip.reset()

        rockHitBullet = self.bullet.collidesGroup(self.rocks)
        if rockHitBullet:
            rockHitBullet.kill()
            if len(self.rockGroup) <= 0:
                self.stop()
                self.carrier.score = self.points
            self.bullet.reset()

        self.points -= 1
        self.lblPoints.text = "%d" % self.points
        if self.points <= 0:
            self.stop()
            score = 0

class Intro(gameEngine.Scene):
    """ introduction.  Simply sets the
        stage
    """
    def __init__(self):
        gameEngine.Scene.__init__(self)
        instructions = gameEngine.MultiLabel()
        instructions.textLines = [
            "You are caught in an asteroid",
            "field. Use your blasters to ",
            "destroy the asteroids.",
            "",
            "Blow them up quickly for more "
            "points."]
        
        instructions.size = (400, 300)
        instructions.fgColor = (0xFF, 0xFF, 0xFF)
        instructions.bgColor = (0, 0, 0)
        instructions.center = (320, 200)
        
        self.button = gameEngine.Button()
        self.button.center = (320, 400)
        self.button.text = "Play"
        
        self.sprites = [instructions, self.button]
        self.setCaption("Asteroids")
        
    def update(self):
        if self.button.clicked:
            self.stop()

class Report(gameEngine.Scene):
    """ reports the player's score, 
        determines if player wants 
        to try again
        score and data passed in 
        carrier object
    """

    def __init__(self, carrier):
        gameEngine.Scene.__init__(self)
        
        self.carrier = carrier
        
        lblPoints = gameEngine.Label()
        lblPoints.center = (320, 240)
        lblPoints.fgColor = (0xFF, 0xFF, 0xFF)
        lblPoints.bgColor = (0, 0, 0) 
        lblPoints.size = (300, 50)
        lblPoints.text = "Score: %d points" % carrier.score
        
        self.btnAgain = gameEngine.Button()
        self.btnAgain.text = "play again"
        self.btnAgain.center = (100, 400)
        
        self.btnQuit = gameEngine.Button()
        self.btnQuit.text = "quit"
        self.btnQuit.center = (540, 400)
        
        self.sprites = [lblPoints, self.btnAgain, self.btnQuit]
        
    def update(self):
        if self.btnAgain.clicked:
            self.carrier.goAgain = True
            self.stop()
        if self.btnQuit.clicked:
            self.carrier.goAgain = False
            self.stop()

class Carrier(object):
    """ an object meant to hold multi-state
        data:
        score: current number of points
        goAgain: boolean continue
    """
    
    def __init__(self, score, goAgain):
        self.score = score
        self.goAgain = goAgain

def main():
    intro = Intro()
    intro.start()

    carrier = Carrier(20000, True)

    while carrier.goAgain:
        game = Game(carrier)
        game.start()
            
        report = Report(carrier)
        report.start()
    
if __name__ == "__main__":
    main()